import bionty.base as bt
import pandas as pd


def test_clo_cellline_inspect_name():
    df = pd.DataFrame(
        index=[
            "253D cell",
            "HEK293",
            "2C1H7 cell",
            "283TAg cell",
            "This cell line does not exist",
        ]
    )

    cl = bt.CellLine(source="clo")
    inspected_df = cl.inspect(df.index, field=cl.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



# Generated by Django 5.0.6 on 2024-05-18 21:18

import django.db.models.deletion
import lnschema_core.models
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0026_artifactcellline_cell_line_ref_is_name_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="artifactcellline",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactcellline",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactcellmarker",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactcellmarker",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactcelltype",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactcelltype",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactdevelopmentalstage",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactdevelopmentalstage",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactdisease",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactdisease",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactethnicity",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactethnicity",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactexperimentalfactor",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactexperimentalfactor",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactgene",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactgene",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactorganism",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactorganism",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactpathway",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactpathway",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactphenotype",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactphenotype",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifactprotein",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifactprotein",
            name="created_by",
        ),
        migrations.RemoveField(
            model_name="artifacttissue",
            name="created_at",
        ),
        migrations.RemoveField(
            model_name="artifacttissue",
            name="created_by",
        ),
        migrations.AlterField(
            model_name="artifactcellline",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="id",
            field=models.BigAutoField(primary_key=True, serialize=False),
        ),
        migrations.CreateModel(
            name="FeatureSetCellMarker",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "cellmarker",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="bionty.cellmarker",
                    ),
                ),
                (
                    "featureset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lnschema_core.featureset",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, lnschema_core.models.LinkORM),
        ),
        migrations.CreateModel(
            name="FeatureSetGene",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "featureset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lnschema_core.featureset",
                    ),
                ),
                (
                    "gene",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="bionty.gene",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, lnschema_core.models.LinkORM),
        ),
        migrations.CreateModel(
            name="FeatureSetPathway",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "featureset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lnschema_core.featureset",
                    ),
                ),
                (
                    "pathway",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="bionty.pathway",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, lnschema_core.models.LinkORM),
        ),
        migrations.CreateModel(
            name="FeatureSetProtein",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "featureset",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="lnschema_core.featureset",
                    ),
                ),
                (
                    "protein",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="bionty.protein",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, lnschema_core.models.LinkORM),
        ),
        migrations.RunSQL(
            """
            INSERT INTO bionty_featuresetgene (featureset_id, gene_id)
            SELECT featureset_id, gene_id
            FROM bionty_gene_feature_sets;
            """
        ),
        migrations.RunSQL(
            """
            INSERT INTO bionty_featuresetprotein (featureset_id, protein_id)
            SELECT featureset_id, protein_id
            FROM bionty_protein_feature_sets;
            """
        ),
        migrations.RunSQL(
            """
            INSERT INTO bionty_featuresetcellmarker (featureset_id, cellmarker_id)
            SELECT featureset_id, cellmarker_id
            FROM bionty_cellmarker_feature_sets;
            """
        ),
        migrations.RunSQL(
            """
            INSERT INTO bionty_featuresetpathway (featureset_id, pathway_id)
            SELECT featureset_id, pathway_id
            FROM bionty_pathway_feature_sets;
            """
        ),
        # Remove the old ManyToMany fields and replace them with the new through models
        migrations.RemoveField(
            model_name="gene",
            name="feature_sets",
        ),
        migrations.AddField(
            model_name="gene",
            name="feature_sets",
            field=models.ManyToManyField(
                through="bionty.FeatureSetGene",
                to="lnschema_core.FeatureSet",
                related_name="genes",
            ),
        ),
        migrations.RemoveField(
            model_name="protein",
            name="feature_sets",
        ),
        migrations.AddField(
            model_name="protein",
            name="feature_sets",
            field=models.ManyToManyField(
                through="bionty.FeatureSetProtein",
                to="lnschema_core.FeatureSet",
                related_name="proteins",
            ),
        ),
        migrations.RemoveField(
            model_name="cellmarker",
            name="feature_sets",
        ),
        migrations.AddField(
            model_name="cellmarker",
            name="feature_sets",
            field=models.ManyToManyField(
                through="bionty.FeatureSetCellMarker",
                to="lnschema_core.FeatureSet",
                related_name="cell_markers",
            ),
        ),
        migrations.RemoveField(
            model_name="pathway",
            name="feature_sets",
        ),
        migrations.AddField(
            model_name="pathway",
            name="feature_sets",
            field=models.ManyToManyField(
                through="bionty.FeatureSetPathway",
                to="lnschema_core.FeatureSet",
                related_name="pathways",
            ),
        ),
        migrations.RenameField(
            model_name="artifactcellline",
            old_name="cell_line",
            new_name="cellline",
        ),
        migrations.RenameField(
            model_name="artifactcellmarker",
            old_name="cell_marker",
            new_name="cellmarker",
        ),
        migrations.RenameField(
            model_name="artifactcelltype",
            old_name="cell_type",
            new_name="celltype",
        ),
        migrations.RenameField(
            model_name="artifactdevelopmentalstage",
            old_name="developmental_stage",
            new_name="developmentalstage",
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Tuple, overload

import numpy as np
from django.db import models
from django.db.models import CASCADE, PROTECT
from lamin_utils import logger
from lnschema_core.models import (
    Artifact,
    CanValidate,
    Feature,
    FeatureSet,
    HasParents,
    LinkORM,
    Record,
    TracksRun,
    TracksUpdates,
)

import bionty.base as bionty_base

from . import ids
from ._bionty import encode_uid, lookup2kwargs

if TYPE_CHECKING:
    from pandas import DataFrame

    from .base import PublicOntology


class StaticReference:
    def __init__(self, source_record: Source) -> None:
        self._source = source_record

    def df(self) -> DataFrame:
        return self._source.dataframe_artifact.load()


class BioRecord(Record, HasParents, CanValidate):
    """Base Record of bionty.

    BioRecord inherits all methods from :class:`~lamindb.core.Record` and provides additional methods
    including :meth:`~bionty.core.BioRecord.public` and :meth:`~bionty.core.BioRecord.from_source`.

    Notes:
        For more info, see tutorials:

        - :doc:`docs:bionty`
        - :doc:`docs:bio-registries`
    """

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        # DB-facing constructor
        if len(args) == len(self._meta.concrete_fields):
            super().__init__(*args, **kwargs)
            return None

        # passing lookup result from bionty, which is a Tuple or List
        if (
            args
            and len(args) == 1
            and isinstance(args[0], (Tuple, List))
            and len(args[0]) > 0
        ):
            if isinstance(args[0], List) and len(args[0]) > 1:
                logger.warning(
                    "multiple lookup/search results are passed, only returning record from the first entry"
                )
            result = lookup2kwargs(self, *args, **kwargs)  # type:ignore
            # exclude "parents" from query arguments
            query_kwargs = {k: v for k, v in result.items() if k != "parents"}
            existing_record = self.__class__.filter(**query_kwargs).one_or_none()
            if existing_record is not None:
                from lamindb._record import init_self_from_db

                init_self_from_db(self, existing_record)
                return None
            else:
                kwargs = result  # result already has encoded id
                args = ()
        # all other cases require encoding the id
        else:
            kwargs = encode_uid(orm=self, kwargs=kwargs)

        # raise error if no organism is passed
        if hasattr(self.__class__, "organism_id"):
            if kwargs.get("organism") is None and kwargs.get("organism_id") is None:
                from .core._settings import settings

                if settings.organism is not None:
                    kwargs["organism"] = settings.organism
                else:
                    raise RuntimeError("please pass a organism!")
            elif kwargs.get("organism") is not None:
                if not isinstance(kwargs.get("organism"), Organism):
                    raise TypeError("organism must be a `bionty.Organism` record")

        # now continue with the user-facing constructor
        # set the direct parents as a private attribute
        # this is a list of strings that store the ontology id
        if "parents" in kwargs:
            parents = kwargs.pop("parents")
            # this checks if we receive a np.ndarray from pandas
            if isinstance(parents, (list, np.ndarray)) and len(parents) > 0:
                if not isinstance(parents[0], str):
                    raise ValueError(
                        "not a valid parents kwarg, got to be list of ontology ids"
                    )
                self._parents = parents

        super().__init__(*args, **kwargs)

    @classmethod
    def list_source(
        cls,
        currently_used: bool | None = None,
        in_db: bool | None = None,
        organism: str | None = None,
    ) -> Source:
        """Default source for the registry.

        Args:
            currently_used: Only returns currently used sources

        Examples:
            >>> bionty.Gene.list_source()
            >>> bionty.Gene.list_source(currently_used=True)
        """
        filters = {}
        if currently_used is not None:
            filters["currently_used"] = currently_used
        if in_db is not None:
            filters["in_db"] = in_db
        if organism is not None:
            filters["organism"] = organism  # type:ignore
        return Source.filter(entity=cls.__get_name_with_schema__(), **filters)

    @classmethod
    def import_from_source(
        cls,
        source: Source | None = None,
        ontology_ids: list[str] | None = None,
        organism: str | Record | None = None,
        ignore_conflicts: bool = True,
        update: bool = False,
    ):
        """Bulk save records from a dataframe.

        Use this method to initialize your registry with public ontology.

        Args:
            ontology_ids: List of ontology ids to save
            organism: Organism record or name
            source: Source record
            ignore_conflicts: Ignore conflicts during bulk create

        Examples:
            >>> bionty.CellType.import_from_source()
        """
        from .core._add_ontology import add_ontology_from_df, check_source_in_db

        if hasattr(cls, "ontology_id"):
            add_ontology_from_df(
                registry=cls,
                ontology_ids=ontology_ids,
                organism=organism,
                source=source,
                ignore_conflicts=ignore_conflicts,
                update=update,
            )
        else:
            import lamindb as ln

            from ._bionty import get_source_record

            public = cls.public(organism=organism, source=source)
            # TODO: consider StaticReference
            source_record = get_source_record(public)  # type:ignore
            df = public.df().reset_index()
            if hasattr(cls, "_ontology_id_field"):
                field = cls._ontology_id_field
            else:
                raise NotImplementedError(
                    f"import_from_source is not implemented for {cls.__name__}"
                )
            records = cls.from_values(
                ontology_ids or df[field],
                field=field,
                organism=organism,
                source=source_record,
            )
            ln.save(records, ignore_conflicts=ignore_conflicts)

            # make sure source.in_db is correctly set based on the DB content
            check_source_in_db(registry=cls, source=source_record, update=update)

    @classmethod
    def public(
        cls,
        organism: str | Record | None = None,
        source: Source | None = None,
    ) -> PublicOntology | StaticReference:
        """The corresponding PublicOntology object.

        Note that the source is auto-configured and tracked via :meth:`bionty.Source`.

        See Also:
            `PublicOntology <https://lamin.ai/docs/public-ontologies>`__

        Examples:
            >>> celltype_pub = bionty.CellType.public()
            >>> celltype_pub
            PublicOntology
            Entity: CellType
            Organism: all
            Source: cl, 2023-04-20
            #terms: 2698
        """
        if isinstance(organism, Organism):
            organism = organism.name

        if source is not None:
            organism = source.organism
            source_name = source.name
            version = source.version
        else:
            from .core._settings import settings

            if hasattr(cls, "organism_id"):
                if organism is None and settings.organism is not None:
                    organism = settings.organism.name
            source_name = None
            version = None

        try:
            return getattr(bionty_base, cls.__name__)(
                organism=organism, source=source_name, version=version
            )
        except (AttributeError, ValueError):
            return StaticReference(source)

    @classmethod
    def from_source(
        cls, *, mute: bool = False, **kwargs
    ) -> BioRecord | list[BioRecord] | None:
        """Create a record or records from source based on a single field value.

        Notes:
            For more info, see tutorial :doc:`docs:bionty`

            Bulk create protein records via :meth:`~docs:lamindb.core.Record.from_values`.

        Examples:
            Create a record by passing a field value:

            >>> record = bionty.Gene.from_source(symbol="TCF7", organism="human")

            Create a record from non-default source:

            >>> source = bionty.Source.filter(entity="CellType", source="cl", version="2022-08-16").one()  # noqa
            >>> record = bionty.CellType.from_source(name="T cell", source=source)

        """
        # non-relationship kwargs
        kv = {
            k: v
            for k, v in kwargs.items()
            if k not in [i.name for i in cls._meta.fields if i.is_relation]
        }
        if len(kv) > 1:
            raise AssertionError(
                "Only one field can be passed to generate record from public reference"
            )
        elif len(kv) == 0:
            return None
        else:
            k = next(iter(kv))
            v = kwargs.pop(k)
            results = cls.from_values([v], field=getattr(cls, k), mute=mute, **kwargs)
            if len(results) == 1:
                return results[0]
            elif len(results) == 0:
                return None
            else:
                return results

    # deprecated
    @classmethod
    def from_public(cls, *args, **kwargs) -> BioRecord | list[BioRecord] | None:
        """Create a record or records from public reference based on a single field value.

        Notes:
            For more info, see tutorial :doc:`docs:bionty`

            Bulk create protein records via :meth:`~docs:lamindb.core.Record.from_values`.

        Examples:
            Create a record by passing a field value:

            >>> record = bionty.Gene.from_public(symbol="TCF7", organism="human")

        """
        logger.warning("`.from_public()` is deprecated, use `.from_source()`!'")
        return cls.from_source(*args, **kwargs)

    def save(self, *args, **kwargs) -> BioRecord:
        """Save the record and its parents recursively."""
        super().save(*args, **kwargs)
        # saving records of parents
        if hasattr(self, "_parents"):
            import lamindb as ln

            # here parents is still a list of ontology ids
            parents = self._parents
            # bulk create parent records
            parents_records = self.from_values(
                parents, self.__class__.ontology_id, source=self.source
            )
            ln.save(parents_records)
            self.parents.set(parents_records)

        return self


class Organism(BioRecord, TracksRun, TracksUpdates):
    """Organism - `NCBI Taxonomy <https://www.ncbi.nlm.nih.gov/taxonomy/>`__, `Ensembl Organism <https://useast.ensembl.org/info/about/species.html>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:organism`.


    Examples:
        >>> record = bionty.Organism.from_source(name="rabbit")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(
        max_length=64, db_index=True, default=None, unique=True
    )
    """Name of a organism, required field."""
    ontology_id: str | None = models.CharField(
        max_length=32, unique=True, db_index=True, null=True, default=None
    )
    """NCBI Taxon ID."""
    scientific_name: str | None = models.CharField(
        max_length=64, db_index=True, unique=True, null=True, default=None
    )
    """Scientific name of a organism."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this organism."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the organism."""
    parents: Organism = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent organism records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="organisms"
    )
    """:class:`~bionty.Source` this record associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactOrganism", related_name="organisms"
    )
    """Artifacts linked to the organism."""

    @overload
    def __init__(
        self,
        name: str,
        taxon_id: str | None,
        scientific_name: str | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Gene(BioRecord, TracksRun, TracksUpdates):
    """Genes - `Ensembl <https://ensembl.org/>`__, `NCBI Gene <https://www.ncbi.nlm.nih.gov/gene/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:gene`.

        Bulk create Gene records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.Gene.from_source(symbol="TCF7", organism="human")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False

    _name_field: str = "symbol"
    _ontology_id_field: str = "ensembl_gene_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=12, default=ids.gene)
    """A universal id (hash of selected field)."""
    symbol: str | None = models.CharField(
        max_length=64, db_index=True, null=True, default=None
    )
    """A unique short form of gene name."""
    stable_id: str | None = models.CharField(
        max_length=64, db_index=True, null=True, default=None, unique=True
    )
    """Stable ID of a gene that doesn't have ensembl_gene_id, e.g. a yeast gene."""
    ensembl_gene_id: str | None = models.CharField(
        max_length=64, db_index=True, null=True, default=None, unique=True
    )
    """Ensembl gene stable ID, in the form ENS[organism prefix][feature type prefix][a unique eleven digit number]."""
    ncbi_gene_ids: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) NCBI Gene IDs that correspond to this Ensembl Gene ID.
    NCBI Gene ID, also known as Entrez Gene ID, in the form of numeric string, 1 to 9 digits.
    """
    biotype: str | None = models.CharField(
        max_length=64, db_index=True, null=True, default=None
    )
    """Type of the gene."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this gene."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the gene."""
    organism: Organism = models.ForeignKey(
        Organism, PROTECT, default=None, related_name="genes"
    )
    """:class:`~bionty.Organism` this gene associates with."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="genes"
    )
    """:class:`~bionty.Source` this gene associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactGene", related_name="genes"
    )
    """Artifacts linked to the gene."""
    feature_sets: FeatureSet = models.ManyToManyField(
        FeatureSet, through="FeatureSetGene", related_name="genes"
    )
    """Featuresets linked to this gene."""

    @overload
    def __init__(
        self,
        symbol: str | None,
        stable_id: str | None,
        ensembl_gene_id: str | None,
        ncbi_gene_ids: str | None,
        biotype: str | None,
        description: str | None,
        synonyms: str | None,
        organism: Organism | None,
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Protein(BioRecord, TracksRun, TracksUpdates):
    """Proteins - `Uniprot <https://www.uniprot.org/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:protein`.

        Bulk create Protein records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.Protein.from_source(name="Synaptotagmin-15B", organism="human")
        >>> record = bionty.Protein.from_source(gene_symbol="SYT15B", organism="human")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False

    _name_field: str = "name"
    _ontology_id_field: str = "uniprotkb_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=12, default=ids.protein)
    """A universal id (hash of selected field)."""
    name: str | None = models.CharField(
        max_length=256, db_index=True, null=True, default=None
    )
    """Unique name of a protein."""
    uniprotkb_id: str | None = models.CharField(
        max_length=10, db_index=True, null=True, default=None, unique=True
    )
    """UniProt protein ID, 6 alphanumeric characters, possibly suffixed by 4 more."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this protein."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the protein."""
    length: int | None = models.BigIntegerField(db_index=True, null=True)
    """Length of the protein sequence."""
    gene_symbol: str | None = models.CharField(
        max_length=256, db_index=True, null=True, default=None
    )
    """The primary gene symbol corresponds to this protein."""
    ensembl_gene_ids: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) Ensembl Gene IDs that correspond to this protein."""
    organism: Organism = models.ForeignKey(
        Organism, PROTECT, default=None, related_name="proteins"
    )
    """:class:`~bionty.Organism` this protein associates with."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="proteins"
    )
    """:class:`~bionty.Source` this protein associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactProtein", related_name="proteins"
    )
    """Artifacts linked to the protein."""
    feature_sets: FeatureSet = models.ManyToManyField(
        FeatureSet, through="FeatureSetProtein", related_name="proteins"
    )
    """Featuresets linked to this protein."""

    @overload
    def __init__(
        self,
        name: str | None,
        uniprotkb_id: str | None,
        synonyms: str | None,
        length: int | None,
        gene_symbol: str | None,
        ensembl_gene_ids: str | None,
        organism: Organism | None,
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class CellMarker(BioRecord, TracksRun, TracksUpdates):
    """Cell markers - `CellMarker <http://xteam.xbio.top/CellMarker>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:cell_marker`.

        Bulk create CellMarker records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.CellMarker.from_source(name="PD1", organism="human")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "organism"),)

    _name_field: str = "name"
    _ontology_id_field: str = "name"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=12, default=ids.cellmarker)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=64, db_index=True)
    """Unique name of the cell marker."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this cell marker."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the cell marker."""
    gene_symbol: str | None = models.CharField(
        max_length=64, db_index=True, null=True, default=None
    )
    """Gene symbol that corresponds to this cell marker."""
    ncbi_gene_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """NCBI gene id that corresponds to this cell marker."""
    uniprotkb_id: str | None = models.CharField(
        max_length=10, db_index=True, null=True, default=None
    )
    """Uniprotkb id that corresponds to this cell marker."""
    organism: Organism = models.ForeignKey(
        Organism, PROTECT, default=None, related_name="cell_markers"
    )
    """:class:`~bionty.Organism` this cell marker associates with."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="cell_markers"
    )
    """:class:`~bionty.Source` this cell marker associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactCellMarker",
        related_name="cell_markers",
    )
    """Artifacts linked to the cell marker."""
    feature_sets: FeatureSet = models.ManyToManyField(
        FeatureSet, through="FeatureSetCellMarker", related_name="cell_markers"
    )
    """Featuresets linked to this cell marker."""

    @overload
    def __init__(
        self,
        name: str,
        synonyms: str | None,
        gene_symbol: str | None,
        ncbi_gene_id: str | None,
        uniprotkb_id: str | None,
        organism: Organism | None,
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Tissue(BioRecord, TracksRun, TracksUpdates):
    """Tissues - `Uberon <http://obophenotype.github.io/uberon/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` :doc:`docs:tissue`.

        Bulk create Tissue records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.Tissue.from_source(name="brain")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the tissue."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the tissue."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of tissue."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this tissue."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the tissue."""
    parents: Tissue = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent tissues records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="tissues"
    )
    """:class:`~bionty.Source` this tissue associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactTissue", related_name="tissues"
    )
    """Artifacts linked to the tissue."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Tissue],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class CellType(BioRecord, TracksRun, TracksUpdates):
    """Cell types - `Cell Ontology <https://obophenotype.github.io/cell-ontology/>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:cell_type`.

        Bulk create CellType records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.CellType.from_source(name="T cell")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the cell type."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the cell type."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of cell type."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this cell type."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the cell type."""
    parents: CellType = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent cell type records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="cell_types"
    )
    """:class:`~bionty.Source` this cell type associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactCellType", related_name="cell_types"
    )
    """Artifacts linked to the cell type."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[CellType],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Disease(BioRecord, TracksRun, TracksUpdates):
    """Diseases - `Mondo <https://mondo.monarchinitiative.org/>`__, `Human Disease <https://disease-ontology.org/>`__.

    Notes:
        Bulk create Disease records via :meth:`~docs:lamindb.core.Record.from_values`.

        For more info, see tutorials: :doc:`docs:disease`.

    Examples:
        >>> record = bionty.Disease.from_source(name="Alzheimer disease")
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the disease."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the disease."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of disease."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this disease."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the disease."""
    parents: Disease = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent disease records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="diseases"
    )
    """:class:`~bionty.Source` this disease associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactDisease", related_name="diseases"
    )
    """Artifacts linked to the disease."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Disease],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class CellLine(BioRecord, TracksRun, TracksUpdates):
    """Cell lines - `Cell Line Ontology <https://github.com/CLO-ontology/CLO>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:cell_line`.

        Bulk create CellLine records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> standard_name = bionty.CellLine.public().standardize(["K562"])[0]
        >>> record = bionty.CellLine.from_source(name=standard_name)
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the cell line."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the cell line."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of cell line."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this cell line."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the cell line."""
    parents: CellLine = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent cell line records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="cell_lines"
    )
    """:class:`~bionty.Source` this cell line associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactCellLine", related_name="cell_lines"
    )
    """Artifacts linked to the cell line."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[CellLine],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Phenotype(BioRecord, TracksRun, TracksUpdates):
    """Phenotypes - `Human Phenotype <https://hpo.jax.org/app/>`__,
    `Phecodes <https://phewascatalog.org/phecodes_icd10>`__,
    `Mammalian Phenotype <http://obofoundry.org/ontology/mp.html>`__,
    `Zebrafish Phenotype <http://obofoundry.org/ontology/zp.html>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:phenotype`.

        Bulk create Phenotype records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.Phenotype.from_source(name="Arachnodactyly")
        >>> record.save()
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the phenotype."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the phenotype."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of phenotype."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this phenotype."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the phenotype."""
    parents: Phenotype = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent phenotype records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="phenotypes"
    )
    """:class:`~bionty.Source` this phenotype associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactPhenotype", related_name="phenotypes"
    )
    """Artifacts linked to the phenotype."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Phenotype],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Pathway(BioRecord, TracksRun, TracksUpdates):
    """Pathways - `Gene Ontology <https://bioportal.bioontology.org/ontologies/GO>`__,
    `Pathway Ontology <https://bioportal.bioontology.org/ontologies/PW>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:pathway`.

        Bulk create Pathway records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.Pathway.from_source(ontology_id="GO:1903353")
        >>> record.save()
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the pathway."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the pathway."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of pathway."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this pathway."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the pathway."""
    parents: Pathway = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent pathway records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="pathways"
    )
    """:class:`~bionty.Source` this pathway associates with."""
    genes: Gene = models.ManyToManyField("Gene", related_name="pathways")
    """Genes that signifies the pathway."""
    feature_sets: FeatureSet = models.ManyToManyField(
        FeatureSet, through="FeatureSetPathway", related_name="pathways"
    )
    """Featuresets linked to the pathway."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, through="ArtifactPathway", related_name="pathways"
    )
    """Artifacts linked to the pathway."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Pathway],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class ExperimentalFactor(BioRecord, TracksRun, TracksUpdates):
    """Experimental factors - `Experimental Factor Ontology <https://www.ebi.ac.uk/ols/ontologies/efo>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:experimental_factor`.

        Bulk create ExperimentalFactor records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> standard_name = bionty.ExperimentalFactor.public().standardize(["scRNA-seq"])
        >>> record = bionty.ExperimentalFactor.from_source(name=standard_name)
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the experimental factor."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the experimental factor."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of experimental factor."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this experimental factor."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the experimental factor."""
    molecule: str | None = models.TextField(null=True, default=None, db_index=True)
    """Molecular experimental factor, parsed from EFO."""
    instrument: str | None = models.TextField(null=True, default=None, db_index=True)
    """Instrument used to measure the experimental factor, parsed from EFO."""
    measurement: str | None = models.TextField(null=True, default=None, db_index=True)
    """Phenotypic experimental factor, parsed from EFO."""
    parents: ExperimentalFactor = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent experimental factor records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="experimental_factors"
    )
    """:class:`~bionty.Source` this experimental_factors associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactExperimentalFactor",
        related_name="experimental_factors",
    )
    """Artifacts linked to the experimental_factors."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[ExperimentalFactor],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class DevelopmentalStage(BioRecord, TracksRun, TracksUpdates):
    """Developmental stages - `Human Developmental Stages <https://github.com/obophenotype/developmental-stage-ontologies/wiki/HsapDv>`__,
    `Mouse Developmental Stages <https://github.com/obophenotype/developmental-stage-ontologies/wiki/MmusDv>`__.  # noqa.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:developmental_stage`.

        Bulk create DevelopmentalStage records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.DevelopmentalStage.from_source(name="neurula stage")
        >>> record.save()
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the developmental stage."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the developmental stage."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of developmental stage."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this developmental stage."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the developmental stage."""
    parents: str | None = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent developmental stage records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="developmental_stages"
    )
    """:class:`~bionty.Source` this developmental stage associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactDevelopmentalStage",
        related_name="developmental_stages",
    )
    """Artifacts linked to the developmental stage."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[DevelopmentalStage],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Ethnicity(BioRecord, TracksRun, TracksUpdates):
    """Ethnicity - `Human Ancestry Ontology <https://github.com/EBISPOT/hancestro>`__.

    Notes:
        For more info, see tutorials :doc:`docs:bio-registries` and :doc:`docs:ethnicity`.

        Bulk create Ethnicity records via :meth:`~docs:lamindb.core.Record.from_values`.

    Examples:
        >>> record = bionty.Ethnicity.from_source(name="European")
        >>> record.save()
    """

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("name", "ontology_id"),)

    _name_field: str = "name"
    _ontology_id_field: str = "ontology_id"

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=8, default=ids.ontology)
    """A universal id (hash of selected field)."""
    name: str = models.CharField(max_length=256, db_index=True)
    """Name of the ethnicity."""
    ontology_id: str | None = models.CharField(
        max_length=32, db_index=True, null=True, default=None
    )
    """Ontology ID of the ethnicity."""
    abbr: str | None = models.CharField(
        max_length=32, db_index=True, unique=True, null=True, default=None
    )
    """A unique abbreviation of ethnicity."""
    synonyms: str | None = models.TextField(null=True, default=None)
    """Bar-separated (|) synonyms that correspond to this ethnicity."""
    description: str | None = models.TextField(null=True, default=None)
    """Description of the ethnicity."""
    parents: Ethnicity = models.ManyToManyField(
        "self", symmetrical=False, related_name="children"
    )
    """Parent ethnicity records."""
    source: Source = models.ForeignKey(
        "Source", PROTECT, null=True, related_name="ethnicities"
    )
    """:class:`~bionty.Source` this ethnicity associates with."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact,
        through="ArtifactEthnicity",
        related_name="ethnicities",
    )
    """Artifacts linked to the ethnicity."""

    @overload
    def __init__(
        self,
        name: str,
        ontology_id: str | None,
        abbr: str | None,
        synonyms: str | None,
        description: str | None,
        parents: list[Ethnicity],
        source: Source | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)


class Source(Record, TracksRun, TracksUpdates):
    """Versions of ontology sources.

    .. warning::

        Do not modify the records unless you know what you are doing!
    """

    _name_field: str = "name"

    class Meta(BioRecord.Meta, TracksRun.Meta, TracksUpdates.Meta):
        abstract = False
        unique_together = (("entity", "name", "organism", "version"),)

    id: int = models.AutoField(primary_key=True)
    """Internal id, valid only in one DB instance."""
    uid: str = models.CharField(unique=True, max_length=4, default=ids.source)
    """A universal id (hash of selected field)."""
    entity: str = models.CharField(max_length=256, db_index=True)
    """Entity class name."""
    organism: str = models.CharField(max_length=64, db_index=True)
    """Organism name, use 'all' if unknown or none applied."""
    name: str = models.CharField(max_length=64, db_index=True)
    """Source name, short form, CURIE prefix for ontologies."""
    version: str = models.CharField(max_length=64, db_index=True)
    """Version of the source."""
    in_db: bool = models.BooleanField(default=False, db_index=True)
    """Whether this ontology has be added to the database."""
    currently_used: bool = models.BooleanField(default=False, db_index=True)
    """Whether this record is currently used."""
    description: str | None = models.TextField(blank=True, db_index=True)
    """Source full name, long form."""
    url: str | None = models.TextField(null=True, default=None)
    """URL of the source file."""
    md5: str | None = models.TextField(null=True, default=None)
    """Hash md5 of the source file."""
    source_website: str | None = models.TextField(null=True, default=None)
    """Website of the source."""
    dataframe_artifact: Artifact = models.ForeignKey(
        Artifact, PROTECT, null=True, default=None, related_name="_source_dataframe_of"
    )
    """Dataframe artifact that corresponds to this source."""
    artifacts: Artifact = models.ManyToManyField(
        Artifact, related_name="_source_artifact_of"
    )
    """Additional files that correspond to this source."""

    @overload
    def __init__(
        self,
        entity: str,
        organism: str,
        name: str,
        version: str,
        currently_used: bool,
        description: str | None,
        url: str | None,
        md5: str | None,
        source_website: str | None,
    ): ...

    @overload
    def __init__(
        self,
        *db_args,
    ): ...

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        kwargs = encode_uid(orm=self, kwargs=kwargs)
        super().__init__(*args, **kwargs)

    def set_as_currently_used(self):
        """Set this record as the currently used source.

        Examples:
            >>> record = bionty.Source.filter(uid="...").one()
            >>> record.set_as_currently_used()
        """
        # set this record as currently used
        self.currently_used = True
        self.save()
        # set all other records as not currently used
        Source.filter(
            entity=self.entity, organism=self.organism, name=self.name
        ).exclude(uid=self.uid).update(currently_used=False)
        logger.success(f"set {self} as currently used")
        logger.warning("please reload your instance to reflect the updates!")


class FeatureSetGene(Record, LinkORM):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    featureset: FeatureSet = models.ForeignKey(
        "lnschema_core.FeatureSet", CASCADE, related_name="+"
    )
    gene: Gene = models.ForeignKey("Gene", PROTECT, related_name="+")


class FeatureSetProtein(Record, LinkORM):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    featureset: FeatureSet = models.ForeignKey(
        "lnschema_core.FeatureSet", CASCADE, related_name="+"
    )
    protein: Protein = models.ForeignKey("Protein", PROTECT, related_name="+")


class FeatureSetCellMarker(Record, LinkORM):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    featureset: FeatureSet = models.ForeignKey(
        "lnschema_core.FeatureSet", CASCADE, related_name="+"
    )
    # follow the .lower() convention in link models
    cellmarker: CellMarker = models.ForeignKey("CellMarker", PROTECT, related_name="+")


class FeatureSetPathway(Record, LinkORM):
    id: int = models.BigAutoField(primary_key=True)
    # follow the .lower() convention in link models
    featureset: FeatureSet = models.ForeignKey(
        "lnschema_core.FeatureSet", CASCADE, related_name="+"
    )
    pathway: Pathway = models.ForeignKey("Pathway", PROTECT, related_name="+")


class ArtifactOrganism(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_organism"
    )
    organism: Organism = models.ForeignKey(
        "Organism", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactorganism"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactGene(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(Artifact, CASCADE, related_name="links_gene")
    gene: Gene = models.ForeignKey("Gene", PROTECT, related_name="links_artifact")
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactgene"
    )
    gene_ref_is_symbol: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_symbol: bool | None = models.BooleanField(null=True, default=None)


class ArtifactProtein(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_protein"
    )
    protein: Protein = models.ForeignKey(
        "Protein", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactprotein"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactCellMarker(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_cell_marker"
    )
    # follow the .lower() convention in link models
    cellmarker: CellMarker = models.ForeignKey(
        "CellMarker", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactcellmarker",
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactTissue(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_tissue"
    )
    tissue: Tissue = models.ForeignKey("Tissue", PROTECT, related_name="links_artifact")
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifacttissue"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactCellType(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_cell_type"
    )
    # follow the .lower() convention in link models
    celltype: CellType = models.ForeignKey(
        "CellType", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactcelltype"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactDisease(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_disease"
    )
    disease: Disease = models.ForeignKey(
        "Disease", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactdisease"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactCellLine(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_cell_line"
    )
    # follow the .lower() convention in link models
    cellline: CellLine = models.ForeignKey(
        "CellLine", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactcellline"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactPhenotype(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_phenotype"
    )
    phenotype: Phenotype = models.ForeignKey(
        "Phenotype", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactphenotype",
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactPathway(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_pathway"
    )
    pathway: Pathway = models.ForeignKey(
        "Pathway", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature, PROTECT, null=True, default=None, related_name="links_artifactpathway"
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactExperimentalFactor(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_experimental_factor"
    )
    experimentalfactor: ExperimentalFactor = models.ForeignKey(
        "ExperimentalFactor", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactexperimentalfactor",
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactDevelopmentalStage(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_developmental_stage"
    )
    # follow the .lower() convention in link models
    developmentalstage: DevelopmentalStage = models.ForeignKey(
        "DevelopmentalStage", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactdevelopmentalstage",
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


class ArtifactEthnicity(Record, LinkORM, TracksRun):
    id: int = models.BigAutoField(primary_key=True)
    artifact: Artifact = models.ForeignKey(
        Artifact, CASCADE, related_name="links_ethnicity"
    )
    ethnicity: Ethnicity = models.ForeignKey(
        "Ethnicity", PROTECT, related_name="links_artifact"
    )
    feature: Feature = models.ForeignKey(
        Feature,
        PROTECT,
        null=True,
        default=None,
        related_name="links_artifactethnicity",
    )
    label_ref_is_name: bool | None = models.BooleanField(null=True, default=None)
    feature_ref_is_name: bool | None = models.BooleanField(null=True, default=None)


# backward compat
Species = Organism
BiontySource = Source
BioRegistry = BioRecord
PublicSource = Source



from typing import Dict, Literal, Optional

import pandas as pd
from lamin_utils import logger

from bionty.base._ontology import Ontology
from bionty.base._public_ontology import PublicOntology
from bionty.base.entities._shared_docstrings import _doc_params, organism_removed


@_doc_params(doc_entities=organism_removed)
class ExperimentalFactor(PublicOntology):
    """Experimental Factor.

    1. Experimental Factor Ontology
    Edits of terms are coordinated and reviewed on:
    https://www.ebi.ac.uk/ols/ontologies/efo

    Args:
        {doc_entities}

    Also see: `bionty.core.PublicOntology <https://docs.lamin.ai/bionty.core.publicontology>`__
    """

    def __init__(
        self,
        organism: Optional[Literal["all"]] = None,
        source: Optional[Literal["efo"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            organism=organism,
            source=source,
            version=version,
            include_id_prefixes={"efo": ["EFO", "http://www.ebi.ac.uk/efo/"]},
            **kwargs,
        )

    def to_pronto(self) -> Ontology:  # type:ignore
        """The Pronto Ontology object.

        See: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html
        """
        self._download_ontology_file(
            localpath=self._local_ontology_path,  # type:ignore
            url=self._url,  # type:ignore
            md5=self._md5,  # type:ignore
        )
        onto = Ontology(
            handle=self._local_ontology_path,  # type:ignore
            prefix="http://www.ebi.ac.uk/efo/",
        )
        # TODO: fix
        onto.__setattr__("efo_to_df", efo_to_df)
        return onto


def _parse_efo_term(
    term_id: str,
    ontology: Ontology,
) -> Dict:
    """Parse readout attributes from EFO."""
    readout_terms = {
        "assay": "OBI:0000070",
        "assay_by_molecule": "EFO:0002772",
        "assay_by_instrument": "EFO:0002773",
        "assay_by_sequencer": "EFO:0003740",
        "measurement": "EFO:0001444",
    }

    def _list_to_str(lst: list):
        if len(lst) == 0:
            return None
        elif len(lst) == 1:
            return lst[0].name  # type: ignore
        else:
            return ";".join([i.name for i in lst])

    def _list_subclasses(ontology: Ontology, term, *, distance=1, with_self=False):
        """Subclasses of a term."""
        termclass = ontology.get_term(term)
        return list(termclass.subclasses(distance=distance, with_self=with_self))

    term = ontology.get_term(term_id)
    superclasses = term.superclasses()

    assay_by_molecule = _list_subclasses(ontology, readout_terms["assay_by_molecule"])
    assay_by_instrument = _list_subclasses(
        ontology, readout_terms["assay_by_instrument"]
    )

    assay_by_sequencer = _list_subclasses(ontology, readout_terms["assay_by_sequencer"])
    measurement = _list_subclasses(ontology, readout_terms["measurement"])

    molecules = [i for i in assay_by_molecule if i in superclasses]
    instruments = [i for i in assay_by_sequencer if i in superclasses]
    if len(instruments) == 0:
        instruments = [i for i in assay_by_instrument if i in superclasses]
    measurements = [i for i in measurement if i in superclasses]

    readout = {
        "ontology_id": term_id,
        "name": term.name,
        "molecule": _list_to_str(molecules),
        "instrument": _list_to_str(instruments),
        "measurement": _list_to_str(measurements),
    }

    return readout


def efo_to_df(
    ontology: Ontology,
    source: str,
    include_id_prefixes: dict,
    prefix: str = "http://www.ebi.ac.uk/efo/",
) -> pd.DataFrame:  # pragma: no cover
    df = ontology.to_df(
        source=source, include_id_prefixes=include_id_prefixes
    ).reset_index()
    df["ontology_id"] = [
        i.replace(prefix, "").replace("_", ":") for i in df["ontology_id"]
    ]
    df["parents"] = [
        [j.replace(prefix, "").replace("_", ":") for j in i] for i in df["parents"]
    ]

    logger.info("parsing EFO terms for the first time will take 6-10min...")
    parsed_results = []
    for term in df["ontology_id"]:
        parsed_results.append(_parse_efo_term(term, ontology))
    df_parsed = pd.DataFrame.from_records(parsed_results)
    df = df.merge(df_parsed).set_index("ontology_id")

    return df



# Generated by Django 5.0.6 on 2024-05-16 15:59

import django.db.models.deletion
import lnschema_core.models
import lnschema_core.users
from django.db import migrations, models

import bionty.ids


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("lnschema_core", "0037_rename_dataset_to_collection"),
    ]

    operations = [
        migrations.CreateModel(
            name="Organism",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True, default=None, max_length=64, unique=True
                    ),
                ),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "scientific_name",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=64,
                        null=True,
                        unique=True,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="organism", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="organism", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_organism",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.organism"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="PublicSource",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.publicsource,
                        max_length=8,
                        unique=True,
                    ),
                ),
                ("entity", models.CharField(db_index=True, max_length=64)),
                ("organism", models.CharField(db_index=True, max_length=64)),
                ("currently_used", models.BooleanField(db_index=True, default=False)),
                ("source", models.CharField(db_index=True, max_length=64)),
                ("source_name", models.TextField(blank=True, db_index=True)),
                ("version", models.CharField(db_index=True, max_length=64)),
                ("url", models.TextField(default=None, null=True)),
                ("md5", models.TextField(default=None, null=True)),
                ("source_website", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_public_sources",
                        to="lnschema_core.user",
                    ),
                ),
            ],
            options={
                "unique_together": {("entity", "source", "organism", "version")},
            },
        ),
        migrations.CreateModel(
            name="Protein",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.protein, max_length=12, unique=True
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True, default=None, max_length=64, null=True
                    ),
                ),
                (
                    "uniprotkb_id",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=10,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("length", models.BigIntegerField(db_index=True, null=True)),
                (
                    "gene_symbol",
                    models.CharField(
                        db_index=True, default=None, max_length=64, null=True
                    ),
                ),
                ("ensembl_gene_ids", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="proteins", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="proteins", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_proteins",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "feature_sets",
                    models.ManyToManyField(
                        related_name="proteins", to="lnschema_core.featureset"
                    ),
                ),
                (
                    "organism",
                    models.ForeignKey(
                        default=None,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="proteins",
                        to="bionty.organism",
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="proteins",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.AddField(
            model_name="organism",
            name="public_source",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="organisms",
                to="bionty.publicsource",
            ),
        ),
        migrations.CreateModel(
            name="Gene",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.gene, max_length=12, unique=True
                    ),
                ),
                (
                    "symbol",
                    models.CharField(
                        db_index=True, default=None, max_length=64, null=True
                    ),
                ),
                (
                    "stable_id",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=64,
                        null=True,
                        unique=True,
                    ),
                ),
                (
                    "ensembl_gene_id",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=64,
                        null=True,
                        unique=True,
                    ),
                ),
                ("ncbi_gene_ids", models.TextField(default=None, null=True)),
                (
                    "biotype",
                    models.CharField(
                        db_index=True, default=None, max_length=64, null=True
                    ),
                ),
                ("description", models.TextField(default=None, null=True)),
                ("synonyms", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="genes", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="genes", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_genes",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "feature_sets",
                    models.ManyToManyField(
                        related_name="genes", to="lnschema_core.featureset"
                    ),
                ),
                (
                    "organism",
                    models.ForeignKey(
                        default=None,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="genes",
                        to="bionty.organism",
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="genes",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="CellMarker",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.cellmarker,
                        max_length=12,
                        unique=True,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True, default=None, max_length=64, unique=True
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                (
                    "gene_symbol",
                    models.CharField(
                        db_index=True, default=None, max_length=64, null=True
                    ),
                ),
                (
                    "ncbi_gene_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "uniprotkb_id",
                    models.CharField(
                        db_index=True, default=None, max_length=10, null=True
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="cell_markers", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="cell_markers", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_cell_markers",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "feature_sets",
                    models.ManyToManyField(
                        related_name="cell_markers", to="lnschema_core.featureset"
                    ),
                ),
                (
                    "organism",
                    models.ForeignKey(
                        default=None,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="cell_markers",
                        to="bionty.organism",
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="cell_markers",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="Phenotype",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="phenotypes", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="phenotypes", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_phenotypes",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.phenotype"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="phenotypes",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="Pathway",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="pathways", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="pathways", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_pathways",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "feature_sets",
                    models.ManyToManyField(
                        related_name="pathways", to="lnschema_core.featureset"
                    ),
                ),
                (
                    "genes",
                    models.ManyToManyField(related_name="pathways", to="bionty.gene"),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.pathway"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="pathways",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="ExperimentalFactor",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("molecule", models.TextField(db_index=True, default=None, null=True)),
                (
                    "instrument",
                    models.TextField(db_index=True, default=None, null=True),
                ),
                (
                    "measurement",
                    models.TextField(db_index=True, default=None, null=True),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="experimental_factors", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="experimental_factors",
                        to="lnschema_core.collection",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_experimental_factors",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.experimentalfactor"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="experimental_factors",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="Ethnicity",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="ethnicities", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="ethnicities", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_ethnicities",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.ethnicity"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="ethnicities",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="Disease",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="diseases", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="diseases", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_diseases",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.disease"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="diseases",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="DevelopmentalStage",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="developmental_stages", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="developmental_stages",
                        to="lnschema_core.collection",
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_developmental_stages",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.developmentalstage"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="developmental_stages",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="CellType",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="cell_types", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="cell_types", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_cell_types",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.celltype"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="cell_types",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="CellLine",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="cell_lines", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="cell_lines", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_cell_lines",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(
                        related_name="children", to="bionty.cellline"
                    ),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="cell_lines",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
        migrations.CreateModel(
            name="Tissue",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "uid",
                    models.CharField(
                        default=bionty.ids.ontology, max_length=8, unique=True
                    ),
                ),
                ("name", models.CharField(db_index=True, max_length=256)),
                (
                    "ontology_id",
                    models.CharField(
                        db_index=True, default=None, max_length=32, null=True
                    ),
                ),
                (
                    "abbr",
                    models.CharField(
                        db_index=True,
                        default=None,
                        max_length=32,
                        null=True,
                        unique=True,
                    ),
                ),
                ("synonyms", models.TextField(default=None, null=True)),
                ("description", models.TextField(default=None, null=True)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True, db_index=True)),
                (
                    "artifacts",
                    models.ManyToManyField(
                        related_name="tissues", to="lnschema_core.artifact"
                    ),
                ),
                (
                    "collections",
                    models.ManyToManyField(
                        related_name="tissues", to="lnschema_core.collection"
                    ),
                ),
                (
                    "created_by",
                    models.ForeignKey(
                        default=lnschema_core.users.current_user_id,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="created_tissues",
                        to="lnschema_core.user",
                    ),
                ),
                (
                    "parents",
                    models.ManyToManyField(related_name="children", to="bionty.tissue"),
                ),
                (
                    "public_source",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="tissues",
                        to="bionty.publicsource",
                    ),
                ),
            ],
            options={
                "unique_together": {("name", "ontology_id")},
            },
            bases=(
                models.Model,
                lnschema_core.models.HasParents,
                lnschema_core.models.CanValidate,
            ),
        ),
    ]



import json
from pathlib import Path
from typing import Literal, Optional

import pandas as pd

from bionty.base._public_ontology import PublicOntology
from bionty.base.dev._io import s3_bionty_assets

from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class BFXPipeline(PublicOntology):
    """Bioinformatics pipelines.

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Optional[Literal["all"]] = None,
        source: Optional[Literal["lamin"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(source=source, version=version, organism=organism, **kwargs)

    def _load_df(self) -> pd.DataFrame:
        localpath = self._local_parquet_path.as_posix().replace(  # type:ignore
            ".parquet", ".json"
        )
        s3_bionty_assets("bfxpipelines.json", Path(localpath))
        with open(localpath) as f:
            data = json.load(f)

        df = pd.DataFrame(data).transpose()
        df.drop("versions", inplace=True, axis=1)
        df.rename(columns={"id": "ontology_id"}, inplace=True)
        df.set_index("ontology_id", inplace=True, drop=True)

        return df

    def df(self) -> pd.DataFrame:
        """Pandas DataFrame of the ontology.

        Returns:
            A Pandas DataFrame of the ontology.

        Examples:
            >>> import bionty.base as bionty_base
            >>> bt.BFXPipeline().df()
        """
        return self._df.set_index("ontology_id")



from typing import Literal, Optional

from bionty.base._public_ontology import PublicOntology

from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class BioSample(PublicOntology):
    """BioSample attributes.

    1. NCBI BioSample Attributes
    https://www.ncbi.nlm.nih.gov/biosample/docs/attributes

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Optional[Literal["all"]] = None,
        source: Optional[Literal["ncbi"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(source=source, version=version, organism=organism, **kwargs)



import bionty.base as bt
import pandas as pd


def test_uniprot_protein_inspect_uniprotkb_id():
    df = pd.DataFrame(
        index=[
            "A0A024QZ08",  # no longer exist in 2024-03 version
            "X6RLV5",
            "X6RM24",
            "A0A024QZQ1",  # no longer exist in 2024-03 version
            "This protein does not exist",
        ]
    )

    pr = bt.Protein(source="uniprot")
    inspected_df = pr.inspect(df.index, pr.uniprotkb_id, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([False, True, True, False, False])

    assert inspect.equals(expected_series)



# Generated by Django 5.0.6 on 2024-05-18 14:14

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0025_artifactcellline_alter_cellline_artifacts_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="artifactcellline",
            name="cell_line_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactcellline",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactcellmarker",
            name="cell_marker_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactcellmarker",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactcelltype",
            name="cell_type_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactcelltype",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactdevelopmentalstage",
            name="developmental_stage_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactdevelopmentalstage",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactdisease",
            name="disease_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactdisease",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactethnicity",
            name="ethnicity_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactethnicity",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactexperimentalfactor",
            name="experimental_factor_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactexperimentalfactor",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactgene",
            name="feature_ref_is_symbol",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactgene",
            name="gene_ref_is_symbol",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactorganism",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactorganism",
            name="organism_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactpathway",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactpathway",
            name="pathway_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactphenotype",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactphenotype",
            name="phenotype_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactprotein",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifactprotein",
            name="protein_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifacttissue",
            name="feature_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AddField(
            model_name="artifacttissue",
            name="tissue_ref_is_name",
            field=models.BooleanField(default=None, null=True),
        ),
        migrations.AlterField(
            model_name="artifactcellline",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="cell_line_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellline",
            name="cell_line",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.cellline",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellline",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactcellline_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="cell_marker_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="cell_marker",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.cellmarker",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcellmarker",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactcellmarker_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="cell_type_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="cell_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.celltype",
            ),
        ),
        migrations.AlterField(
            model_name="artifactcelltype",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactcelltype_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="developmental_stage_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="developmental_stage",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.developmentalstage",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdevelopmentalstage",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactdevelopmentalstage_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="disease_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="disease",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.disease",
            ),
        ),
        migrations.AlterField(
            model_name="artifactdisease",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactdisease_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="ethnicity_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="ethnicity",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.ethnicity",
            ),
        ),
        migrations.AlterField(
            model_name="artifactethnicity",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactethnicity_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="experimental_factor_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="experimental_factor",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.experimentalfactor",
            ),
        ),
        migrations.AlterField(
            model_name="artifactexperimentalfactor",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactexperimentalfactor_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="gene_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactgene_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactgene",
            name="gene",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.gene",
            ),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="organism_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactorganism_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactorganism",
            name="organism",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.organism",
            ),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="pathway_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactpathway_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactpathway",
            name="pathway",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.pathway",
            ),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="phenotype_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactphenotype_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactphenotype",
            name="phenotype",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.phenotype",
            ),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="protein_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifactprotein_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifactprotein",
            name="protein",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.protein",
            ),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="artifact",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="tissue_links",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="feature",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifacttissue_links",
                to="lnschema_core.feature",
            ),
        ),
        migrations.AlterField(
            model_name="artifacttissue",
            name="tissue",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="artifact_links",
                to="bionty.tissue",
            ),
        ),
    ]



from pathlib import Path
from typing import Tuple

import pytest
from bionty.base.dev._md5 import calculate_md5, verify_md5

CURRENT_DIR = Path(__file__).parent


@pytest.fixture(scope="module")
def file_fixture() -> Tuple[str, str]:  # type: ignore
    file_path = "test_file.txt"
    with open(file_path, "w") as f:
        f.write("Hello, world!")
    expected_md5 = calculate_md5(file_path)
    yield file_path, expected_md5
    Path(file_path).unlink()


def test_verify_md5_with_matching_md5(file_fixture):
    file_path, expected_md5 = file_fixture
    assert verify_md5(file_path, expected_md5)


def test_verify_md5_with_non_matching_md5(file_fixture):
    file_path, _ = file_fixture
    assert not verify_md5(file_path, "0123456789abcdef0123456789abcdef")



import base64
import hashlib
import json
import os
import secrets
import string
from pathlib import Path
from typing import Dict, Optional, Union

from github import Github
from rich.progress import track

BASE_BFX_PIPELINES_PATH = "./scripts/bfxpipelines_info"


def base62(n_char: int) -> str:
    """Like nanoid without hyphen and underscore."""
    alphabet = string.digits + string.ascii_letters.swapcase()
    id = "".join(secrets.choice(alphabet) for i in range(n_char))
    return id


def to_b64_str(bstr: bytes) -> str:
    b64 = base64.urlsafe_b64encode(bstr).decode().strip("=")
    return b64


def hash_str(s: str) -> str:
    bstr = s.encode("utf-8")
    # as we're truncating at a short length, we choose md5 over sha512
    return to_b64_str(hashlib.md5(bstr).digest())


def hash_id(input_id: Optional[str] = None, *, n_char: int) -> str:
    if input_id is None:
        return base62(n_char=n_char)
    else:
        return hash_str(input_id)[:n_char].replace("_", "0").replace("-", "0")


def generate_nf_core_pipelines_info() -> None:
    """Generates a json file that contains all required pipelines information by querying the nf-core Github org."""
    gh_login = Github(os.getenv("GITHUB_TOKEN"))
    nf_core_org = gh_login.get_organization("nf-core")
    blacklist = ["cookiecutter", "tools"]
    nf_core_pipelines = {}

    for repo in track(
        nf_core_org.get_repos(),
        description="Fetching information from nf-core repositories...",
    ):
        if "pipeline" in list(repo.get_topics()):
            if repo.name in blacklist:
                continue

            for version in repo.get_releases():
                actual_version = (
                    version.tag_name if len(version.tag_name) >= 1 else "pre-release"
                )
                pipeline_name = f"{repo.name} v{actual_version}"
                underscore_pipeline_name = (
                    pipeline_name.replace(" ", "_").replace(".", "_").replace("-", "_")
                )

                nf_core_pipelines[underscore_pipeline_name] = {
                    "id": hash_id(pipeline_name, n_char=12),
                    "name": f"nf-core {pipeline_name}",
                    "versions": actual_version,
                    "reference": repo.url,
                }

    with open(f"{BASE_BFX_PIPELINES_PATH}/nf_core_pipelines.json", "w") as f:
        json_data = json.dumps(nf_core_pipelines, indent=4)
        f.write(json_data)


def merge_json_files(pipelines_folder_path: Union[str, Path], output_path: str) -> None:
    """Merge all JSON files in a folder and write the merged data to a new JSON file.

    Args:
        pipelines_folder_path: Path to the folder containing the JSON files.
        output_path: Path to the output JSON file.
    """
    pipelines_folder_path = Path(pipelines_folder_path)
    file_paths = list(pipelines_folder_path.glob("*.json"))

    pipeline_json: Dict = {}

    for file_path in file_paths:
        with open(file_path) as f:
            if not str(file_path).endswith("bfxpipelines.json"):
                pipelines_info = json.load(f)
                pipeline_json = {**pipeline_json, **pipelines_info}

    with open(output_path, "w") as f:
        json.dump(pipeline_json, f, indent=4)


generate_nf_core_pipelines_info()
merge_json_files(
    pipelines_folder_path=BASE_BFX_PIPELINES_PATH,
    output_path=f"{BASE_BFX_PIPELINES_PATH}/bfxpipelines.json",
)



import warnings
from pathlib import Path
from typing import BinaryIO, Dict, List, Optional, Union

import pandas as pd
import pronto


class Ontology(pronto.Ontology):
    """Interface with ontologies via pronto.

    Also see: https://pronto.readthedocs.io/en/stable/api/pronto.Ontology.html

    Args:
        handle: Path to an ontology source file.
        import_depth: The maximum depth of imports to resolve in the ontology tree.
        timeout: The timeout in seconds to use when performing network I/O.
        threads: The number of threads to use when parsing.
        url: The url of the ontology.
        prefix: Dev only -> prefix for get_term.
    """

    def __init__(
        self,
        handle: Union[str, Path, BinaryIO, None] = None,
        import_depth: int = -1,
        timeout: int = 100,
        threads: Optional[int] = None,
        prefix: str = "",
    ) -> None:
        self._prefix = prefix
        warnings.filterwarnings("ignore", category=pronto.warnings.ProntoWarning)
        super().__init__(
            handle=handle, import_depth=import_depth, timeout=timeout, threads=threads
        )

    def get_term(self, term):
        """Search an ontology by its id."""
        try:
            return super().get_term(term)
        except KeyError:
            return super().get_term(f"{self._prefix}{term.replace(':', '_')}")

    def to_df(
        self,
        source: Optional[str] = None,
        include_id_prefixes: Optional[Dict[str, List[str]]] = None,
    ):
        """Convert pronto.Ontology to a DataFrame with columns id, name, parents."""

        def filter_include_id_prefixes(terms: pronto.ontology._OntologyTerms):
            if include_id_prefixes and source in list(include_id_prefixes.keys()):
                return list(
                    filter(
                        lambda val: any(
                            val.id.startswith(prefix)
                            for prefix in include_id_prefixes[source]  # type: ignore
                        ),
                        terms,
                    )
                )
            else:
                return terms

        if source is not None:
            prefix_list = (
                include_id_prefixes.get(source)
                if include_id_prefixes is not None
                else None
            )
        else:
            prefix_list = None

        filtered_terms = filter_include_id_prefixes(self.terms())

        df_values = []
        for term in filtered_terms:
            # skip terms without id or name
            if (not term.id) or (not term.name):
                continue

            # term definition text
            definition = None if term.definition is None else term.definition.title()

            # concatenate synonyms into a string
            synonyms = "|".join(
                [i.description for i in term.synonyms if i.scope == "EXACT"]
            )
            if len(synonyms) == 0:
                synonyms = None  # type:ignore

            # get 1st degree parents as a list
            if prefix_list is not None:
                superclasses = [
                    s.id
                    for s in term.superclasses(distance=1, with_self=False).to_set()
                    if s.id.startswith(tuple(prefix_list))
                ]
            else:
                superclasses = [
                    s.id
                    for s in term.superclasses(distance=1, with_self=False).to_set()
                ]

            df_values.append((term.id, term.name, definition, synonyms, superclasses))

        df = pd.DataFrame(
            df_values,
            columns=["ontology_id", "name", "definition", "synonyms", "parents"],
        ).set_index("ontology_id")

        # needed to avoid erroring in .lookup()
        df["name"] = df["name"].fillna("")

        return df



def _upload_ontology_artifacts(instance: str, lamin_user: str, lamin_api_key: str):
    import bionty_base as bt
    import lamindb as ln
    from bionty.base._public_ontology import encode_filenames
    from bionty.base._settings import settings

    ln.setup.login(lamin_user, key=lamin_api_key)
    ln.setup.load(instance)

    queryset = ln.File.select().all()

    files = []
    for entity, row in bt.display_available_sources().iterrows():
        parquet_filename, ontology_filename = encode_filenames(
            organism=row.organism, source=row.source, version=row.version, entity=entity
        )
        if entity == "Organism" or row.url.startswith("s3://bionty-assets"):
            continue

        if not queryset.filter(key=parquet_filename).exists():
            local_parquet_filename = settings.dynamicdir / parquet_filename
            if not local_parquet_filename.exists():
                try:
                    getattr(bt, entity)(
                        organism=row.organism, source=row.source, version=row.version
                    )
                except Exception:  # for renamed classes
                    continue
            file = ln.File(local_parquet_filename, key=parquet_filename)
            files.append(file)

        if not queryset.filter(key=ontology_filename).exists():
            local_ontology_filename = settings.dynamicdir / ontology_filename
            if not local_ontology_filename.exists():
                getattr(bt, entity)(
                    organism=row.organism, source=row.source, version=row.version
                ).to_pronto()
            file = ln.File(local_ontology_filename, key=ontology_filename)
            files.append(file)
    if len(files) > 0:
        ln.save()

    ln.setup.close()



from filelock import FileLock  # type: ignore

from ._settings import settings
from .dev._handle_sources import (
    create_currently_used_sources_yaml,
    create_or_update_sources_local_yaml,
)


def sync_sources():
    # Make this code safe when running bionty from multiple processes
    with FileLock(settings.versionsdir / "bionty_base.lock"):
        create_or_update_sources_local_yaml(overwrite=False)
        # always generate a new CURRENT_SOURCES file
        create_currently_used_sources_yaml(overwrite=True)



from typing import Type

import pandas as pd
from lamin_utils import logger
from lamindb_setup.core._setup_bionty_sources import RENAME
from lnschema_core.models import Record

import bionty.base as bionty_base


def sync_all_sources_to_latest():
    """Sync up the Source registry with the latest available sources.

    Examples:
        >>> from bionty.core import sync_all_sources_to_latest
        >>> sync_all_sources_to_latest()
    """
    import lamindb as ln

    from bionty.models import Source

    try:
        ln.settings.creation.search_names = False
        records = Source.filter().all()
        df_sources = bionty_base.display_available_sources().reset_index()
        for _, row in df_sources.iterrows():
            kwargs = row.to_dict()
            for db_name, base_name in RENAME.items():
                if base_name in kwargs:
                    kwargs[db_name] = kwargs.pop(base_name)
            if not kwargs["entity"].startswith("bionty."):
                kwargs["entity"] = "bionty." + kwargs["entity"]
            record = records.filter(
                name=kwargs["name"],
                version=kwargs["version"],
                entity=kwargs["entity"],
                organism=kwargs["organism"],
            ).all()
            if len(record) == 0:
                record = Source(**kwargs)
                record.save()
                logger.success(f"added {record}")
            else:
                # update metadata fields
                record.update(**kwargs)
                logger.success(f"updated {record.one()}")
        logger.info("setting the latest version as currently_used...")
        set_latest_sources_as_currently_used()
        logger.success("synced up Source registry with the latest available sources")
        logger.warning("please reload your instance to reflect the updates!")
    finally:
        ln.settings.creation.search_names = True


def set_latest_sources_as_currently_used():
    """Set the currently_used column to True for the latest version of each source.

    Examples:
        >>> from bionty.core import set_latest_sources_as_currently_used
        >>> set_latest_sources_as_currently_used()
    """
    from bionty.models import Source

    records = Source.filter().all()
    df = records.df()
    for (entity, organism), df_group in df.groupby(["entity", "organism"]):
        latest_uid = df_group.sort_values("version", ascending=False).uid.iloc[0]
        records.filter(uid=latest_uid).update(currently_used=True)
        records.filter(entity=entity, organism=organism).exclude(uid=latest_uid).update(
            currently_used=False
        )


def filter_bionty_df_columns(
    model: Type[Record], public_ontology: bionty_base.PublicOntology
) -> pd.DataFrame:
    """Filter columns of public ontology to match the model fields."""
    bionty_df = pd.DataFrame()
    if public_ontology is not None:
        model_field_names = {i.name for i in model._meta.fields}
        # parents needs to be added here as relationships aren't in fields
        model_field_names.add("parents")
        bionty_df = _prepare_bionty_df(model, public_ontology.df().reset_index())
        bionty_df = bionty_df.loc[:, bionty_df.columns.isin(model_field_names)]
    return bionty_df


def _prepare_bionty_df(model: type[Record], bionty_df: pd.DataFrame):
    """Prepare the bionty DataFrame to match the model fields."""
    if model.__get_name_with_schema__() == "bionty.Gene":
        # groupby ensembl_gene_id and concat ncbi_gene_ids
        groupby_id_col = (
            "ensembl_gene_id" if "ensembl_gene_id" in bionty_df else "stable_id"
        )
        bionty_df.drop(
            columns=["hgnc_id", "mgi_id", "index"], errors="ignore", inplace=True
        )
        bionty_df.drop_duplicates([groupby_id_col, "ncbi_gene_id"], inplace=True)
        bionty_df["ncbi_gene_id"] = bionty_df["ncbi_gene_id"].fillna("")
        bionty_df = (
            bionty_df.groupby(groupby_id_col)
            .agg(
                {
                    "symbol": "first",
                    "ncbi_gene_id": "|".join,
                    "biotype": "first",
                    "description": "first",
                    "synonyms": "first",
                }
            )
            .reset_index()
        )
        bionty_df.rename(columns={"ncbi_gene_id": "ncbi_gene_ids"}, inplace=True)

    # rename definition to description for the bionty registry in db
    bionty_df.rename(columns={"definition": "description"}, inplace=True)
    return bionty_df



import bionty.base as bt
import pandas as pd


def test_uberon_tissue_inspect_ontology_id():
    df = pd.DataFrame(
        index=[
            "UBERON:0000000",
            "UBERON:0000005",
            "UBERON:8600001",
            "UBERON:8600002",
            "This tissue does not exist",
        ]
    )

    ts = bt.Tissue(source="uberon")
    inspected_df = ts.inspect(df.index, ts.ontology_id, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)



from pathlib import Path
from tempfile import NamedTemporaryFile, gettempdir

from bionty.base.dev._handle_sources import parse_sources_yaml
from bionty.base.dev._io import s3_bionty_assets, url_download
from bionty.base.dev._md5 import calculate_md5
from rich import print

df = parse_sources_yaml()

for _, row in df.iterrows():
    url = row["url"]
    md5 = row["md5"]
    if not md5:
        if not url.startswith("s3"):
            with NamedTemporaryFile() as temp_file:
                tmp_file_path = temp_file.name
                url_download(url, tmp_file_path)
                md5 = calculate_md5(tmp_file_path)
        elif url.startswith("s3"):
            file_name = url.removeprefix("s3://bionty-assets/")
            local_path = Path(f"{gettempdir()}/{file_name}")
            s3_bionty_assets(
                file_name, localpath=local_path, assets_base_url="s3://bionty-assets"
            )
            try:
                md5 = calculate_md5(local_path)
            except FileNotFoundError:
                print(f"[bold red]URL {url} could not be downloaded. Is it on S3?")
                continue
        else:
            raise ValueError(f"URL type for: {url} not recognized")

        print(f"[bold blue]URL: [green]{url} [blue]has md5: [green]{md5}")



from pathlib import Path

import bionty.base as bt
from bionty.base._settings import settings
from bionty.base.dev._handle_sources import LAMINDB_INSTANCE_LOADED


def test_loaded_lamindb():
    import shutil

    shutil.copyfile(
        settings.current_sources.as_posix(), settings.lamindb_sources.as_posix()
    )

    lnenv = Path.home() / ".lamin/current_instance.env"
    with open(lnenv.as_posix(), "w") as f:
        f.write("schema_str=bionty")

    assert LAMINDB_INSTANCE_LOADED()

    ct = bt.CellType(version="2022-08-16")
    assert ct.source == "cl"

    lnenv.unlink()
    assert not LAMINDB_INSTANCE_LOADED()



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ObjectDoesNotExist
from lamin_utils import logger
from lnschema_core.models import Record

import bionty.base as bionty_base

from . import ids

if TYPE_CHECKING:
    from types import ModuleType


def create_or_get_organism_record(
    organism: str | Record | None, orm: Record
) -> Record | None:
    # return None if an Record doesn't have organism field
    organism_record = None
    if hasattr(orm, "organism_id"):
        # using global setting of organism
        from .core._settings import settings

        if organism is None and settings.organism is not None:
            logger.debug(f"using global setting organism = {settings.organism.name}")
            return settings.organism

        if isinstance(organism, Record):
            organism_record = organism
        elif isinstance(organism, str):
            from .models import Organism

            try:
                # existing organism record
                organism_record = Organism.objects.get(name=organism)
            except ObjectDoesNotExist:
                try:
                    # create a organism record from bionty reference
                    organism_record = Organism.from_source(name=organism)
                    # link the organism record to the default bionty source
                    organism_record.source = get_source_record(bionty_base.Organism())  # type:ignore
                    organism_record.save()  # type:ignore
                except KeyError:
                    # no such organism is found in bionty reference
                    organism_record = None

        if organism_record is None:
            raise AssertionError(
                f"{orm.__name__} requires to specify a organism name via `organism=` or `bionty.settings.organism=`!"
            )

    return organism_record


# TODO: consider private sources
def get_source_record(public_ontology: bionty_base.PublicOntology):
    import bionty

    from .models import Source

    bionty_models = list_biorecord_models(bionty)
    entity_name = public_ontology.__class__.__name__
    if entity_name in bionty_models:
        entity_name = f"bionty.{entity_name}"
    kwargs = {
        "entity": entity_name,
        "organism": public_ontology.organism,
        "name": public_ontology.source,
        "version": public_ontology.version,
    }

    source_record = Source.objects.filter(**kwargs).get()
    return source_record


def list_biorecord_models(schema_module: ModuleType):
    """List all BioRecord models in a given schema module."""
    import inspect

    from .models import BioRecord

    return [
        attr
        for attr in dir(schema_module.models)
        if inspect.isclass(getattr(schema_module.models, attr))
        and issubclass(getattr(schema_module.models, attr), BioRecord)
    ]


def encode_uid(orm: Record, kwargs: dict):
    if kwargs.get("uid") is not None:
        # if uid is passed, no encoding is needed
        return kwargs
    try:
        name = orm.__name__.lower()
    except AttributeError:
        name = orm.__class__.__name__.lower()
    ontology = False
    str_to_encode = None
    if name == "gene":
        str_to_encode = kwargs.get("ensembl_gene_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("stable_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("symbol")
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError("must provide ensembl_gene_id, stable_id or symbol")
    elif name == "protein":
        str_to_encode = kwargs.get("uniprotkb_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("name")
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError("must provide uniprotkb_id or name")
    elif name == "cellmarker":
        str_to_encode = kwargs.get("name")
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError("must provide name")
    elif name == "source":
        str_to_encode = f'{kwargs.get("entity", "")}{kwargs.get("name", "")}{kwargs.get("organism", "")}{kwargs.get("version", "")}'
    else:
        str_to_encode = kwargs.get("ontology_id")
        if str_to_encode is None or str_to_encode == "":
            str_to_encode = kwargs.get("name")
        if str_to_encode is None or str_to_encode == "":
            raise AssertionError("must provide ontology_id or name")
        ontology = True

    if str_to_encode is not None and len(str_to_encode) > 0:
        if ontology:
            id_encoder = ids.ontology
        else:
            try:
                id_encoder = getattr(ids, name)
            except Exception:
                return kwargs
        kwargs["uid"] = id_encoder(str_to_encode)
    return kwargs


def lookup2kwargs(orm: Record, *args, **kwargs) -> dict:
    """Pass bionty search/lookup results."""
    arg = args[0]
    if isinstance(arg, tuple):
        bionty_kwargs = arg._asdict()  # type:ignore
    else:
        bionty_kwargs = arg[0]._asdict()

    if len(bionty_kwargs) > 0:
        import bionty.base as bionty_base

        # add organism and source
        organism_record = create_or_get_organism_record(
            orm=orm.__class__, organism=kwargs.get("organism")
        )
        if organism_record is not None:
            bionty_kwargs["organism"] = organism_record
        public_ontology = getattr(bionty_base, orm.__class__.__name__)(
            organism=organism_record.name if organism_record is not None else None
        )
        bionty_kwargs["source"] = get_source_record(public_ontology)

        model_field_names = {i.name for i in orm._meta.fields}
        model_field_names.add("parents")
        bionty_kwargs = {
            k: v for k, v in bionty_kwargs.items() if k in model_field_names
        }
    return encode_uid(orm=orm, kwargs=bionty_kwargs)


# backward compat
create_or_get_species_record = create_or_get_organism_record



from typing import Literal, Optional

from bionty.base._public_ontology import PublicOntology

from ._shared_docstrings import _doc_params, doc_entites


@_doc_params(doc_entities=doc_entites)
class Drug(PublicOntology):
    """Drug ontologies.

    1. DRON
    Edits of terms are coordinated and reviewed on:
    https://bioportal.bioontology.org/ontologies/DRON/

    Args:
        {doc_entities}
    """

    def __init__(
        self,
        organism: Optional[Literal["all"]] = None,
        source: Optional[Literal["dron"]] = None,
        version: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            source=source,
            version=version,
            organism=organism,
            include_id_prefixes={"dron": ["DRON"]},
            **kwargs,
        )



# Generated by Django 5.1 on 2024-08-01 12:34

import django.db.models.deletion
from django.db import migrations, models

import bionty
import bionty.ids
from bionty._bionty import encode_uid, list_biorecord_models


def prepend_bionty_to_entity(apps, schema_editor):
    bionty_models = list_biorecord_models(bionty)
    Source = apps.get_model("bionty", "Source")
    for source in Source.objects.all():
        if source.entity in bionty_models and not source.entity.startswith("bionty."):
            # append bionty to entity
            source.entity = f"bionty.{source.entity}"
            # re-encode uid
            source.uid = encode_uid(
                Source,
                {
                    "entity": source.entity,
                    "name": source.name,
                    "organism": source.organism,
                    "version": source.version,
                },
            )["uid"]
            source.save()


class Migration(migrations.Migration):
    dependencies = [
        ("bionty", "0031_alter_cellmarker_name_and_more"),
        (
            "lnschema_core",
            "0056_rename_ulabel_ref_is_name_artifactulabel_label_ref_is_name_and_more",
        ),
    ]

    operations = [
        migrations.RenameField(
            model_name="source",
            old_name="source_name",
            new_name="description",
        ),
        migrations.RenameField(
            model_name="source",
            old_name="source",
            new_name="name",
        ),
        migrations.RenameField(
            model_name="source",
            old_name="df",
            new_name="dataframe_artifact",
        ),
        migrations.AlterField(
            model_name="source",
            name="dataframe_artifact",
            field=models.ForeignKey(
                default=None,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="source_dataframe_of",
                to="lnschema_core.artifact",
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="artifacts",
            field=models.ManyToManyField(
                related_name="source_artifact_of", to="lnschema_core.artifact"
            ),
        ),
        migrations.AlterField(
            model_name="source",
            name="entity",
            field=models.CharField(db_index=True, max_length=256),
        ),
        migrations.AlterField(
            model_name="source",
            name="uid",
            field=models.CharField(
                default=bionty.ids.source, max_length=4, unique=True
            ),
        ),
        migrations.RunPython(prepend_bionty_to_entity),
    ]



import bionty.base as bt
import pandas as pd


def test_hp_phenotype_inspect_name():
    df = pd.DataFrame(
        index=[
            "Specific learning disability",
            "Dystonia",
            "Cerebral hemorrhage",
            "Slurred speech",
            "This phenotype does not exist",
        ]
    )

    pt = bt.Phenotype(source="hp")
    inspected_df = pt.inspect(df.index, field=pt.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_mp_phenotype_inspect_name():
    df = pd.DataFrame(
        index=[
            "decreased ovary apoptosis",
            "abnormal Ebner's gland morphology",
            "abnormal lacrimal sac morphology",
            "abnormal nictitating membrane morphology",
            "This phenotype does not exist",
        ]
    )

    pt = bt.Phenotype(source="mp")
    inspected_df = pt.inspect(df.index, field=pt.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_zp_phenotype_inspect_name():
    df = pd.DataFrame(
        index=[
            "somitogenesis disrupted, abnormal",
            "somite specification disrupted, abnormal",
            "liver has extra parts of type collagen trimer liver, abnormal",
            "neuromast hair cell normal process quality apoptotic process, abnormal",
            "This phenotype does not exist",
        ]
    )

    pt = bt.Phenotype(source="zp")
    inspected_df = pt.inspect(df.index, field=pt.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_phe_phenotype_inspect_name():
    df = pd.DataFrame(
        index=[
            "Intestinal infection due to C. difficile",
            "Sepsis and SIRS",
            "Systemic inflammatory response syndrome (SIRS)",
            "Septic shock",
            "This phenotype does not exist",
        ]
    )

    pt = bt.Phenotype(source="phe")
    inspected_df = pt.inspect(df.index, field=pt.name, return_df=True)

    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)


def test_path_phenotype():
    df = pd.DataFrame(
        index=[
            "nocturnal",
            "male",
            "female",
            "mobility",
            "This phenotype does not exist",
        ]
    )

    pt = bt.Phenotype(source="pato")
    inspected_df = pt.inspect(df.index, field=pt.name, return_df=True)
    inspect = inspected_df["__validated__"].reset_index(drop=True)
    expected_series = pd.Series([True, True, True, True, False])

    assert inspect.equals(expected_series)
